/**********************************************************************
  Blink for RTOS  [Jun.05.2016] by T.Kohki
  
  Turns on an LED2 on for one second, then off for one second, repeatedly.
  LED1 is blinking when LED2 is turn off, after several second.
 **********************************************************************/

#include "RTOS.h"                         //  RTOSの利用を定義

const int led1 = 13;
const int led2 = 7;

const uint32_t  Interval1 = 100000L;
const uint32_t  Interval2 = 1000000L;

volatile uint8_t HandleOfTask1, HandleOfTask2;

/**********************************************************************
 *  the setup function runs once when you press reset or power the board
 **********************************************************************/
void setup() {
    //  100 000us毎に登録されたtask1を繰り返し実行する場合の例
    //  task1の後ろに()はつけません。()をつけるとエラーになります
    HandleOfTask1 = Rtos.Loading(0, Interval1, task1);

    //  500 000us毎に登録されたtask2を繰り返し実行する場合の例
    //  task2の後ろに()はつけません。()をつけるとエラーになります
    HandleOfTask2 = Rtos.Loading(0, Interval2, task2);

    //  Task ID = 1 (loop関数)のタスク起動間隔を100ms周期に変更する例
    //  Rtos.Timer(HandleOfTask1, 100000L);

    // initialize digital pins both 13 and 7 as an output.
    pinMode(led1, OUTPUT);
    pinMode(led2, OUTPUT);
}

/**********************************************************************
 *  Task1  100 000us毎に１回実行されるタスクの登録例
 **********************************************************************/
void task1(void) {
    static char	a;
	
    digitalWrite(led1, a);    // LED1 blink
    a = ~a;
  
    //  処理が終わったらタスク実行の終了を宣言します。
    //  Rtos.Suspend(RTOS_CURRENTTASK) を指定しない場合、連続実行となります。
    //  繰り返し実行が指定されていないタスクの場合、再起動せず、実行終了になります。
    //  タイマーによる繰り返し実行は、優先度の低いタイマー割り込みの様な動きになり
    //  ますので、一定間隔でのタスク実行がある程度保証されます。
    Rtos.Suspend(RTOS_CURRENTTASK);

    //  Sleep命令は、一定時間実行を休止する扱いになりますので、タイマーによる
    //  繰り返し実行が指定されていないタスクを指定時間後に再実行する場合に使われます。
    //  大抵のRTOSでは、Sleepを使うケースが多いと思いますが、
    //  タイマーによる繰り返し実行とは異なり、一定間隔でのタスク実行は保証されません。
    //  Sleep命令は、タイムアウト処理の記述に適しています。
    //Rtos.Sleep(RTOS_CURRENTTASK, Interval1);
}

/**********************************************************************
 *  Task2  1 000 000us毎に１回実行されるタスクの登録例
 **********************************************************************/
void task2(void) {
    static char  a, b, c=0;
  
    digitalWrite(led2, a);    // LED blink
    a = ~a;
    if (a != 0) { //  LED2 が OFF の時、Task1 の実行を許可
        if (b == 0) {
//            Rtos.Sleep(HandleOfTask1, Interval1);  // このやり方では指定時間後に、１ショット実行するだけ
//            Rtos.Timer(HandleOfTask1, Interval1);  // このやり方でも再開できますが、カウンターは初期化されます

            //  タイマー設定、カウンターの状態を変化させずにTask1の繰り返し実行を再開する
            Rtos.Resume(HandleOfTask1);        // Task1の実行を再開（Dormant -> Waiting）
            Rtos.Continue(HandleOfTask1);      // Timer設定による繰り返し実行を許可
            b = 1;
        }
    }
    else {        //  LED2 が 点灯しているとき、Task1 の実行を停止
        if (c > 3) {
            Rtos.End(HandleOfTask1);                //  Task1 を停止
            //  Unloadingを実行すると、関数登録情報が削除され、Resumeで再開できません。
            //  Rtos.Unloading(HandleOfTask1);      //  Task1 の登録を解除
        }
        else {
            c++;                                    //  最初は Task1 を制御しない
        }
        b = 0;
    }
  
    //  処理が終わったらタスク実行の終了を宣言します。
    //  Rtos.Suspend(RTOS_CURRENTTASK) を指定しない場合、連続実行となります。
    Rtos.Suspend(RTOS_CURRENTTASK);
}

/**********************************************************************
 *  the loop function runs over and over again forever
 **********************************************************************/
void loop() {
	
  //  RTOSの実行
  Rtos.Control();
}
