/**********************************************************************
 *	Simplified RTOS for Arduino Ver. 1.43 [Jun.05.2016] by T.Kohki
 *
 *	■Abouthe the Simplified RTOS
 *	  1, ノンプリエンティブマルチタスク機能を提供するライブラリです。
 *		Non-preemptive multitasking feature is provided by this library.
 *	  2, タスクの一定時間間隔での繰り返し実行機能（タイマー割り込みと類似）を提供します。
 *		Providing periodically execution feature which is similarly as timer interrupt.
 *	  3, Task番号は１番から始めます。
 *		0番は、自分自身または新規登録とします。
 *		Task Number is assigned from 1.
 *		Task Number 0 is own(current executing) task or create new task.
 *	  4, 計時分解能は Timer 0 割り込みの時間に依存します
 *		Resolution of time count is depend on period of Timer 0 interrupt.
 *	  5, 省電力には対応していません。
 *		This library does not provide power saving mode.
 *	  6, C++専用で作られているため、C言語からの呼び出しには対応していません。
 *		It cannot call this library from C, because this library is made by C++.
 *
 *	■Simplified RTOS の動作
 *	  １、タスク優先順位という概念はありません。タスクの登録順番にスキャン実行されます
 *		CPU時間のタイムアウト処理機能はないので、タスク処理は可能な限り細分化し、
 *		複雑なタスクはサブタスクに分解して実装してください。
 *	  ２、timer割り込みに相当する機能が提供されますので、timer 割り込み処理を使わなくても、
 *		簡単な設定で、ほぼ一定周期間隔でユーザープログラムを自動起動することができます。
 *	  ３, WDTはサポートしていません
 *		別のライブラリを使って実装してください。
 *
 *	■ライブラリの使い方
 *	  1, setup() にRTOSタスクを一つずつ登録します。
 *			例：1,000,000us毎にPlcUserProg1()を呼び出す記述
 *			Rtos.Loading(0,1000000, PlcUserProg1);
 *			Arduino 標準ライブラリのMultipleに似た機能です。
 *	  2, loop() にRTOSの制御プログラムを記述します
 *			Rtos.Control();
 *
 *	  3, 登録できるタスク数の上限数は8としていますが、RTOS.hの設定により変更可能です。
 *
 *	■ ライセンス条件
 *	  Released under the MIT license
 *	  http://opensource.org/licenses/mit-license.php 
 *
 *	  licenses/MIT_license - Open Source Group Japan Wiki
 *	  - Open Source Group Japan - SourceForge.JP
 *	  http://sourceforge.jp/projects/opensource/wiki/licenses%2FMIT_license
 *
 *************************************************************************/
#include <stdint.h>
#ifdef ARDUINO
  #include "Arduino.h"
#endif
#include "RTOS.h"

/*************************************************************************
 *	Global Value
 *************************************************************************/
RTOS	Rtos = RTOS();							//	デフォルトインスタンスの定義

/*************************************************************************
 *	RTOSのデフォルトコンストラクタ
 *	インスタンスの初期化（ヒープを使った配列確保は移植性が悪くなるため行わない）
 *************************************************************************/
RTOS::RTOS(void)
{
	uint8_t	id;
	
	NumOfTask= 0;								//	登録されているタスクの数
	
	for(id = 0; id < RTOS_MAXTASK; id++) {
		Flag[id] = RTOS_UNLOADED;				//	未登録として初期化
		TimerSet[id] = 0;						//	Setting Value
		Counter[id] = 0;						//	Current Value
		Task[id] = Rtos_Hook;					//	dummy vector
	}

	Delta_us	= 0L;							//	Differential Time[us]
	PresentSysTick	= (micros() & 0x00ffffffL);	//
	PreviousTime	= PresentSysTick;

	//	スキャンタイム計時処理の初期化
	scan_max = 0;
	scan_avg = 0;
}

/*************************************************************************
 *	AdjustCount	[Jun.27.2015]
 *	micros()関数の返値について、
 *	16MHzでは、   64L/(F_CPU/1000000L) = 4.0usなので問題ないが、
 *	20MHzの場合は、64L/(F_CPU/1000000L) = 3.2usになる。
 *	しかし、数値は符号無し32bit整数で扱われるため、3usに丸められ、3.2 / 3.0 ＝ 1.0667の
 *	丸め誤差が発生し、タイミングが延びてしまう。
 *	この関数は、あらかじめ設定値を補正し、適正な計時処理が行われるために用いられる。
 *	補正値の計算 1.0000 = 10000
 *************************************************************************/
uint32_t RTOS::AdjustCount(uint32_t us)
{
//	F_CPU が 1, 2, 4, 8, 16MHz の場合、補正不要です
#if (F_CPU==1000000L) || (F_CPU==2000000L) || (F_CPU==4000000L) || (F_CPU==8000000L) || (F_CPU==16000000L)
	return (us);
#elif (RTOS_ADJUST==1L) || (RTOS_ADJUST==0L)
	//	補正不要と設定した場合
	//	・1msのSysTickTimerを利用し、[us]計時できる LPC1114 の場合は、補正不要です
	//	・CMT0とCMT1を併用し、[us]計時できる RX63 マイコンの場合は、補正不要です
	return (us);
#else
	//	下記の補正演算は、F_CPUが上記以外で正確なSysTick Timerを持たないAVRマイコンの場合に必要です
	//	この補正処理は64bit演算を行うため、コードサイズ、実行時間ともに長い。
	uint64_t	FreqAdjust;		//クロック補正値

	FreqAdjust = (64L/(F_CPU/1000000L));						//	整数値を計算
	FreqAdjust = (640000L/(F_CPU/1000000L)) / FreqAdjust;		//	固定小数点補正値
	return ((uint32_t)(((uint64_t)us * 10000L) / FreqAdjust));	//	数値補正
#endif
}

/*************************************************************************
 *	Loading the Task Setting	[Jun.27.2015]
 *		id : 登録対象のタスク番号
 *			0   :追加登録
 *			1以上：指定タスクの動作設定をの強制変更
 *		us = RTOS_EVENTDRIVEN: 定時実行無し
 *		us = RTOS_DOEACHSCAN : 毎スキャン実行
 *		us >= 2 : 指定時間経過後にタスク実行を開始する
 *	※CPU負荷を下げるため、定時実行タスクのは0.1ms以上単位で定義する
 *		登録したい関数名
 *************************************************************************/
uint8_t RTOS::Loading(uint8_t TaskId, uint32_t us, void (*f)()) 
{
	uint8_t		idnum;

	//	制限値を超えてタスク登録を行おうとした場合
	if (NumOfTask >= RTOS_MAXTASK) return (RTOS_INVALID);

	//	タスクIDの変換
	if (TaskId == RTOS_CURRENTTASK) {
		idnum = NumOfTask++;						// タスクの新規登録
	}
	else {
		idnum = TaskId-1;							// 指定Idのタスク状態を再設定する
	}

	//	タスク状態の確認
	if ((Flag[idnum] & RTOS_UNLOADED) != RTOS_UNLOADED) return (RTOS_INVALID);

	//	タスク状態の初期化
	if (us == RTOS_EVENTDRIVEN) {
		Flag[idnum] = RTOS_DORMANT;					//	定時実行しない場合
		TimerSet[idnum] = RTOS_EVENTDRIVEN;			//	Setting Value
		}
	else {
		if (us == RTOS_DOEACHSCAN) {				//	毎スキャン実行の場合
			Flag[idnum] = RTOS_READY | RTOS_WAITING | RTOS_CONTINUE;
			TimerSet[idnum] = RTOS_DOEACHSCAN;		//	Setting Value
		}
		else {										//	周期実行の場合
			Flag[idnum] = RTOS_WAITING | RTOS_CONTINUE;
			if (AdjustCount(us) < RTOS_MINCOUNT) {	//	Timer 0 分解能以上とする
				TimerSet[idnum] = RTOS_MINCOUNT;
			}
			else {
				TimerSet[idnum] = AdjustCount(us);	//	Setting Value
			}
		}
	}
	Task[idnum] = f;								//	関数ポインタの登録
	Counter[idnum]  = RTOS_MINCOUNT / 2;			//	カウンター初期値

	//	変更したタスク番号を返す
	return (idnum+1);
}

/*************************************************************************
 *	UnLoading the Registered Task
 *		id : 登録対象のタスク番号
 *			0   :現在実行中のタスク
 *			1以上：指定タスクの動作設定をの強制変更
 *************************************************************************/
uint8_t RTOS::Unloading(uint8_t TaskId)
{
	uint8_t	idnum;
	
	//	制限値を超えてタスク登録を行おうとした場合
	if (TaskId >= RTOS_MAXTASK) return (RTOS_INVALID);

	//	タスクIDの変換
	if (TaskId == RTOS_CURRENTTASK) {
		idnum = Id;								// 自身のId番号を設定する
	}
	else {
		idnum = TaskId-1;						// 指定Id番号を設定する
	}

	//	タスク状態の初期化
	TimerSet[idnum] = RTOS_EVENTDRIVEN;			//	Setting Value
	Counter[idnum]  = RTOS_MINCOUNT / 2;		//	カウンター初期値
	Flag[idnum] = RTOS_UNLOADED;				//	Change Status Flag into Unloaded
	Task[idnum] = Rtos_Hook;					//	関数ポインタの初期化

	//	変更したタスク番号を返す
	return (idnum+1);
}

/*************************************************************************
 *	Change Timer Setting of Task
 *		us = RTOS_EVENTDRIVEN: Event Driven (Non periodical running)
 *		us = RTOS_DOEACHSCAN : Execute by each scan
 *		us >= 2 : Execute periodically at the defined time
 *	Note) Periodically time is recommended more than 0.1ms
 *************************************************************************/
uint8_t RTOS::Timer(uint8_t TaskId, uint32_t us)
{
	uint8_t	idnum;
	
	//	制限範囲外のタスク番号が指定された場合
	if (TaskId > NumOfTask) return	(RTOS_INVALID);

	if (TaskId == RTOS_CURRENTTASK) {
		idnum = Id;									//	自身のId番号を設定する
	}
	else {
		idnum = TaskId - 1;							//	指定Id番号を設定する
	}

	if (us == RTOS_EVENTDRIVEN) {
		Flag[idnum] = RTOS_DORMANT;             //	周期実行しない場合
		TimerSet[idnum] = RTOS_EVENTDRIVEN;     //	Setting Value
		}
	else {
		if (us == RTOS_DOEACHSCAN) {				//	毎スキャン実行の場合
			Flag[idnum] = RTOS_READY | RTOS_WAITING | RTOS_CONTINUE;
			TimerSet[idnum] = RTOS_DOEACHSCAN;		//	Setting Value
		}
		else {										//	周期実行の場合
			Flag[idnum] = RTOS_WAITING | RTOS_CONTINUE;
			if (AdjustCount(us) < RTOS_MINCOUNT) {	//	Timer 0 分解能以上とする
				TimerSet[idnum] = RTOS_MINCOUNT;
			}
			else {
				TimerSet[idnum] = AdjustCount(us);	//	Setting Value
			}
		}
	}
	Counter[idnum]  = RTOS_MINCOUNT / 2;			//	カウンター初期値
	return (Flag[idnum]);
}

/*************************************************************************
 *	Begin(TaskId);
 *		指定タスクをReadyに遷移させます
 *		タイマーは変更されません
 *************************************************************************/
void RTOS::Begin(uint8_t TaskId)
{
	//	制限範囲外のタスク番号が指定された場合
	if (TaskId > NumOfTask) return;

	if (TaskId == RTOS_CURRENTTASK) {
		Flag[Id] |= RTOS_READY;					// Ready Flagをセット
	}
	else {
		Flag[TaskId-1] |= RTOS_READY;			// Ready Flagをセット
	}
};

/*************************************************************************
 *	End(ID taskid);
 *		指定タスクをDormantに遷移させる。
 *		定時実行タスクは、その動作を停止します
 *************************************************************************/
void RTOS::End(uint8_t TaskId)
{
	//	制限範囲外のタスク番号が指定された場合
	if (TaskId > NumOfTask) return;

	if (TaskId == RTOS_CURRENTTASK) {
		Flag[Id] = RTOS_DORMANT;
	}
	else {
		Flag[TaskId-1] = RTOS_DORMANT;
	}
}

/*************************************************************************
 *	Sleep(TaskId, TimeOut);
 *		指定タスクをWaitingに遷移させ、指定時間(TimeOut[us])後に起床する。
 *		TimeOut = 0 : 即時実行
 *		Waiting を Enable, Ready を Disable
 *		Timer の時刻分解能は Timer 0 の割り込み周期で決まります
 *		周期実行タスクの場合は、Event Driven に変更されます。
 *************************************************************************/
void RTOS::Sleep(uint8_t TaskId, uint32_t TimeOut)
{
	int8_t	idnum;

	//	制限範囲外のタスク番号が指定された場合
	if (TaskId > NumOfTask) return;

	if (TaskId == RTOS_CURRENTTASK) {
		idnum = Id;
	}
	else {
		idnum = TaskId - 1;
	}
	Flag[idnum] 	= RTOS_WAITING;
	#ifdef Arduino_h
		#if ((F_CPU != 8000000L) && (F_CPU != 16000000L))
			TimerSet[idnum] = AdjustCount(TimeOut);	//	Setting Value
		#endif
	#else
			TimerSet[idnum] = TimeOut;				//	Setting Value
	#endif
	Counter[idnum]  = RTOS_MINCOUNT / 2;			//	カウンター初期値
}

/*************************************************************************
 *	Suspend(ID_taskid);
 *		指定タスクをReadyからWaitingに遷移させる。
 *************************************************************************/
void RTOS::Suspend(uint8_t TaskId)
{
	//	制限範囲外のタスク番号が指定された場合
	if (TaskId > NumOfTask) return;

	if (TaskId == RTOS_CURRENTTASK) {
		Flag[Id] = Flag[Id] & (~RTOS_READY);				// taskflagを終了設定
	}
	else {
		Flag[TaskId-1] = Flag[TaskId-1] & (~RTOS_READY);	// taskflagを終了設定
	}
};

/*************************************************************************
 *	Resume(ID taskid);
 *		指定タスクをDormantからWaitingに遷移させる。
 *************************************************************************/
void RTOS::Resume(uint8_t TaskId)
{
	//	制限範囲外のタスク番号が指定された場合
	if (TaskId > NumOfTask) return;

	if (TaskId == RTOS_CURRENTTASK) {
		if (Flag[Id] == RTOS_DORMANT)
			Flag[Id] = RTOS_WAITING;				// taskflagを終了設定
	}
	else {
		if (Flag[TaskId-1] == RTOS_DORMANT)
			Flag[TaskId-1] = RTOS_WAITING;			// taskflagを終了設定
	}
};

/*************************************************************************
 *  Continue(ID taskid);
 *    指定タスクのContinue FlagをEnableする。
 *    但し、Timer設定が０の場合は、繰り返し実行にはならない
 *************************************************************************/
void RTOS::Continue(uint8_t TaskId)
{
    //  制限範囲外のタスク番号が指定された場合
    if (TaskId > NumOfTask) return;

    //  Timer設定が０以外の場合にContinue Flagを設定する
    if (TaskId == RTOS_CURRENTTASK) {
        if (TimerSet[Id] != RTOS_EVENTDRIVEN) {
            Flag[Id] |= RTOS_CONTINUE;
        }
    }
    else {
        if (TimerSet[TaskId-1] != RTOS_EVENTDRIVEN) {
            Flag[TaskId-1] |= RTOS_CONTINUE;
        }
    }
};

/*************************************************************************
 *	status(ID taskid)
 *		タスクの状態を参照し、status flag の状態を返す
 *		Task番号0は現在実行中のタスクと定義される
 *************************************************************************/
int8_t RTOS::status(uint8_t TaskId)
{
	//	制限範囲外のタスク番号が指定された場合
	if (TaskId > NumOfTask) return(-2);

	if (TaskId == RTOS_CURRENTTASK) {
		return (Flag[Id]);
	}
	else {
		return (Flag[TaskId-1]);
	}
};

/*************************************************************************
 *	Id()
 *		自タスクの番号を返す
 *************************************************************************/
uint8_t RTOS::IdNum(void)
{
	return (Id+1);
};

/*************************************************************************
 *	Scantime_Max		[Mar.16.2016]
 *		スキャンタイムの最大値を返す
 *************************************************************************/
uint32_t	RTOS::ScanTimeMax(void)
{
	return (scan_max);
};

/*************************************************************************
 *	Scantime_Avg		[Mar.16.2016]
 *		スキャンタイムの平均値を返す
 *************************************************************************/
uint32_t	RTOS::ScanTimeAvg(void)
{
	return (scan_avg);
};

/*************************************************************************
 *	Hook
 *		何もしないタスク
 *************************************************************************/
void Rtos_Hook(void)
{
	//RTOS::End(RTOS_CURRENTTASK);
};

/**************************************************************************
 *	Task Dispatcher	[Mar.07.2015] by T.Kohki	基本動作確認済み
 *	計時分解能は 1000000/F_CPU*64/256 (2.048ms at 8MHz) です。
 *	各タスク処理は、極力 10[ms] 以内に終了させてください
 **************************************************************************/
void RTOS::Control(void)
{
	//	Waiting処理
	if ((Flag[Id] & RTOS_WAITING) == RTOS_WAITING) {
		Counter[Id] += Delta_us;
		if (Counter[Id] >= TimerSet[Id]) {				//	Timer up判定
			if ((Flag[Id] & RTOS_CONTINUE) == RTOS_CONTINUE) {	//	一定周期実行の場合
				Flag[Id] |= RTOS_READY;					//	Enable Ready
				}
			else {
				Flag[Id] = RTOS_READY;					//	Event Driven の場合
			}
		}
	}

	//	Ready処理
	if ((Flag[Id] & RTOS_READY) == RTOS_READY) {
		if (TimerSet[Id] > 0) {							//	Event Driven の場合は何もしない
			if (Counter[Id] >= TimerSet[Id]) {
				Counter[Id] = Counter[Id] % TimerSet[Id];	//	カウンタ値の調整
			}
		}
		(*Task[Id])();									//	指定タスクを実行
	}

	//	タスク実行の優先順位はなく、実行対象のIdはスキャン毎に更新されます。
	//	１スキャンが 65.536sec 以内なら下記計時処理は正常に機能しますが、
	//	長時間CPU時間を占有するとWDTエラーが発生しますので、状態遷移の考え方により
	//	処理を分割し、複数回の呼び出しで一連の処理が終わるようにコーディングしてください。
	Id++;
	if (Id >= NumOfTask) {
		Id = 0;
		//	Calculate the Delta_us
		//	Delta_us is calculated by SysTickTimer value
		PresentSysTick = micros();

    #ifdef SYSTICK_US_COUNT
    //  RXマイコン（SysTickTimer.h利用）
		if (PresentSysTick >= PreviousTime) {
			Delta_us = (PresentSysTick - PreviousTime);
		}
		else {
			Delta_us = ((PresentSysTick + SYSTICK_US_COUNT) - PreviousTime);
		}
		//	上記は下記コードと同じ演算結果になりますが、5byte程短い（除算を使わないメリットがある）
    //		Delta_us = ((PresentSysTick + SYSTICK_US_COUNT) - PreviousTime) % SYSTICK_US_COUNT;	
    #else
    //  Arduino
    Delta_us = (PresentSysTick - PreviousTime + 0x1000000L) & 0xffffffL;
    #endif
		PreviousTime = PresentSysTick;
		
		//	スキャンタイム[us]最大値の更新処理
		interval += Delta_us;
		if (interval > RTOS_INTERVAL) {
			interval	-= RTOS_INTERVAL;			//	計時調整
			scan_max	= scan_max2;				//	最大スキャンタイム記録結果の更新
			scan_max2	= 0;						//	特定区間のスキャンタイム最大値をクリア
		}
		//
		if (scan_max2 < Delta_us) {					//	最大スキャンタイムの更新
			scan_max2 = Delta_us;
		}
		
		//	スキャンタイム平均値[ns]の計算(1.000[us]精度)
		scan_avg = (scan_avg * 999L / 1000L + Delta_us);
	}
}

/**************************************************************************
 *	■来歴管理
 *  Ver. 1.43 [Jun.05.2016] by T.Kohki
 *  1,  Arduinoで動作するように書き換え（IDEでコンパイルできるように修正)
 *  2,  Continue命令を追加
 *  3,  MITライセンスで一般公開
 *  
 *	Ver. 1.4 [Mar.30.2016] by T.Kohki
 *	1,	RX63 マイコン用に書き換え（CS+でコンパイルできるように改版)
 *	2,	CMT0とCMT1を使ったSystickTimer.cppとの併用で、正確な[us]計時が行えます。
 *	3,	micros()のカウンタリセットに伴うDelta_us計時不具合を修正
 *	4,	RTOSの実行状況確認方法を変更
 *	5,	カウンター値の調整方法が不適切で、WDTエラーの原因になっていたため修正
 *
 *	Ver. 1.32 [Jul.26.2015] by T.Kohki
 *	1,	Suspend処理はReady Flagをクリアする形に修正
 *		なぜか、計時処理が安定するようになった。
 *
 *	Ver. 1.31 [Jul.12.2015] by T.Kohki
 *	1,	LPC1114FN28版は正確な1ms SysTickTimer割り込み処理を使うため、micors()の
 *		計時処理補正が不要になったため、補正処理ソフトを変更
 *	2,  計時補正処理は、F_CPUが2, 4, 8, 16MHz以外のAVRマイコンで行う
 *		ただし、計時補正処理はコードサイズ、実行時間共に長いため、補正不要指定できるように変更
 *
 *	Ver. 1.30 [Jun.27.2015] by T.Kohki
 *	1,	最小実行時間と計時処理の方法を見直し
 *	2,	F_CPUが8MHz, 16MHz以外だとmicros()関数で丸め誤差が発生するため、設定値の補正を追加した
 *	3,	周期実行フラグを新たに定義
 *
 *	Ver. 1.20 [May.23.2015] by T.Kohki
 *	1,	状態遷移フラグの管理方法、コマンド名称を見直し
 *
 *	Ver. 1.11 [May.09.2015] by T.Kohki
 *	1,	タスクの呼び出し方法を１スキャンで全タスク実行から１スキャン１タスクに変更
 *	2,	Arduinoのライブラリとして利用できるような形に変更
 *	3,	Task unloadに対応
 *	4,	タイマーカウントの補正用として計時調整値(RTOS_ADJUST)を採用. 8MHzで130us程度
 *	5,	8MHzにて最小スキャンタイム at 1 taskは、約1ms
 *
 *	Ver. 1.10 [May.06.2015] by T.Kohki
 *	1, Arduino用初期版として製作し、基本動作が確認できた。
 *
 *	Ver. 1.00 [May.06.2013] by T.Kohki
 *	1, AVR StudioのC言語版
 *
 **************************************************************************/
